// Copyright (c) 2009, 2010 Paul Pogonyshev.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.


// This is a benchmark, we don't want assertions to slow anything down.
#ifndef NDEBUG
#  define NDEBUG
#endif

#include "benchmark/config.hpp"

#include <algorithm>
#include <cstddef>
#include <ctime>
#include <iostream>
#include <sstream>
#include <vector>

#include <mct/closed-hash-set.hpp>

#if HAVE_UNORDERED_SET
#  include <unordered_set>
#endif

#if HAVE_TR1_UNORDERED_SET
#  include <tr1/unordered_set>
#endif

#if HAVE_BOOST_UNORDERED_SET_HPP
#  include <boost/unordered_set.hpp>
#endif

#if HAVE_GOOGLE_DENSE_HASH_SET
#  include <google/dense_hash_set>
#endif


using namespace std;


namespace
{

  struct error_exit
  { };


  template <size_t approximate_size>
  struct insert_erase_int_benchmark
  {
    template <typename set_type>
    static  size_t
    do_run (set_type& set, size_t loop_size)
    {
      // Level playing field a bit.
      set.max_load_factor (0.7);
      set.rehash (approximate_size);

      size_t  size_sum = 0;

      for (size_t k = 0; k < loop_size; ++k)
        {
          // This is tested to give various intermediate sizes up to 'approximate_size' or
          // a little smaller.  Surely biased, but should be good enough for a benchmark.
          const int  number = ((k * (1 + 2 * k)) % approximate_size);
          if (k % (7 * approximate_size) <= 4 * approximate_size)
            set.insert (number);
          else
            set.erase (number);

          size_sum += set.size ();
        }

      return size_sum;
    }
  };


  struct benchmark_loop
  {
    size_t           loop_size;
    size_t           num_tries;
    size_t           check_sum;
    vector <double>  used_times;


    explicit
    benchmark_loop (size_t loop_size, size_t num_tries)
      : loop_size (loop_size),
        num_tries (num_tries)
    { }

    template <typename implementation_type, typename benchmark_type>
    void
    do_run ()
    {
      for (size_t pass = 0; pass < num_tries; ++pass)
        {
          implementation_type  implementation;

          clock_t  start     = clock ();
          size_t   check_sum = benchmark_type::do_run (implementation, loop_size);
          clock_t  finish    = clock ();
          double   used_time = static_cast <double> (finish - start) / CLOCKS_PER_SEC;

          if (pass > 0 && check_sum != this->check_sum)
            {
              cerr << "\n*** error ***: check sum differs between passes: "
                   << this->check_sum << " and " << check_sum << "\n";
              throw error_exit ();
            }

          this->check_sum = check_sum;
          used_times.push_back (used_time);

          if (num_tries > 1)
            cerr << used_time << (pass + 1 < num_tries ? ' ' : '\n');
        }
    }

    template <typename implementation_type>
    void
    run (const string& benchmark_name)
    {
      if (benchmark_name == "ie-int-small")
        do_run <implementation_type, insert_erase_int_benchmark <100> > ();
      else if (benchmark_name == "ie-int-large")
        do_run <implementation_type, insert_erase_int_benchmark <10000> > ();
      else if (benchmark_name == "ie-int-huge")
        do_run <implementation_type, insert_erase_int_benchmark <1000000> > ();
      else
        {
          cerr << "Unknown benchmark '" << benchmark_name << "'\n";
          throw error_exit ();
        }
    }

    double
    used_time () const
    {
      return *min_element (used_times.begin (), used_times.end ());
    }
  };


  // MCT's internal "huge prime multiplication" trick seems to give it a solid benefit in
  // cases of bad hash functions.  So, let's use something not as bad for benchmarking and
  // comparing with other implementations.  You can test that this indeed disadvantages
  // MCT on average by removing 31337 below.
  struct rehash_int : MCT_HASH_NAMESPACE::hash <int>
  {
    size_t
    operator() (int value) const
    {
      return 31337 * MCT_HASH_NAMESPACE::hash <int>::operator() (value);
    }
  };


  template <typename type, typename hasher>
  class mct_closed_hash_set : public mct::closed_hash_set <type, hasher>
  {
    typedef  mct::closed_hash_set <type, hasher>  base_type;

  public:

    ~mct_closed_hash_set ()
    {
      // This can be enabled with SCons command line; see 'variables.template'.
#   if MCT_ENABLE_DEBUGGING
      typename base_type::statistics  stats (base_type::collect_statistics ());

      cerr << "MCT closed_hash_table statistics:\n";
      cerr << "  number of buckets:   " << base_type::bucket_count () << "\n";
      cerr << "  debris ratio:        " << stats.debris_ratio << "\n";
      cerr << "  present item lookup: " << stats.avg_present_lookup
           << " (max. " << stats.max_present_lookup << ")\n";
      cerr << "  absent item lookup:  " << stats.avg_absent_lookup
           << " (max. " << stats.max_absent_lookup << ")\n";
      cerr << "  used memory:         " << base_type::used_memory () / 1024.0 << " KB\n";
#   endif
    }
  };


#if HAVE_GOOGLE_DENSE_HASH_SET

  class int_google_dense_hash_set : public google::dense_hash_set <int, rehash_int>
  {
  public:

    int_google_dense_hash_set ()
    {
      // Benchmarks are written to never use negative values.
      set_empty_key (-1);
      set_deleted_key (-2);
    }
  };

#endif


  void
  run_benchmark (const string& implementation, const string& benchmark_name,
                 size_t loop_size, size_t num_tries, bool internal = false)
  {
    if (implementation == "all")
      {
        run_benchmark ("mct", benchmark_name, loop_size, num_tries, true);
#     if HAVE_UNORDERED_SET
        run_benchmark ("std", benchmark_name, loop_size, num_tries, true);
#     endif
#     if HAVE_TR1_UNORDERED_SET
        run_benchmark ("tr1", benchmark_name, loop_size, num_tries, true);
#     endif
#     if HAVE_BOOST_UNORDERED_SET_HPP
        run_benchmark ("boost", benchmark_name, loop_size, num_tries, true);
#     endif
#     if HAVE_GOOGLE_DENSE_HASH_SET
        run_benchmark ("google", benchmark_name, loop_size, num_tries, true);
#     endif
        return;
      }

    if (internal)
      cerr << "Benchmarking implementation '" << implementation << "'\n";

    benchmark_loop  loop (loop_size, num_tries);

    if (implementation == "mct")
      loop.run <mct_closed_hash_set <int, rehash_int> > (benchmark_name);

# if HAVE_UNORDERED_SET
    else if (implementation == "std")
      loop.run <std::unordered_set <int, rehash_int> > (benchmark_name);
# endif

# if HAVE_TR1_UNORDERED_SET
    else if (implementation == "tr1")
      loop.run <std::tr1::unordered_set <int, rehash_int> > (benchmark_name);
# endif

# if HAVE_BOOST_UNORDERED_SET_HPP
    else if (implementation == "boost")
      loop.run <boost::unordered_set <int, rehash_int> > (benchmark_name);
# endif

# if HAVE_GOOGLE_DENSE_HASH_SET
    else if (implementation == "google")
      loop.run <int_google_dense_hash_set> (benchmark_name);
# endif

    else
      {
        cerr << "Unsupported set implementation '" << implementation << "'\n";
        throw error_exit ();
      }

    cerr << "Used time: " << loop.used_time () << " s\n";
    cerr << "Check sum: " << loop.check_sum << "\n";
    if (internal)
      cerr << "\n";
  }

  void
  do_parse (const string& input, size_t& result)
  {
    istringstream  parser (input);
    parser >> result;

    if (parser.fail ())
      {
        cerr << "Cannot parse '" << input << "' as integer\n";
        throw error_exit ();
      }
  }

}


int
main (int argc, char** argv)
{
  size_t  loop_size = 100000000;
  size_t  num_tries = 1;

  if (argc < 3 || 5 < argc)
    {
      cerr << "Usage: " << argv[0] << " IMPLEMENTATION BENCHMARK-NAME [LOOP-SIZE [NUM-TRIES]]\n";

      cerr << "\nSupported implementations are:\n";
      cerr << "  mct -- mct::closed_hash_set <...>\n";
#   if HAVE_UNORDERED_SET
      cerr << "  std -- std::unordered_set <...>\n";
#   endif
#   if HAVE_TR1_UNORDERED_SET
      cerr << "  tr1 -- std::tr1::unordered_set <...>\n";
#   endif
#   if HAVE_BOOST_UNORDERED_SET_HPP
      cerr << "  boost -- boost::unordered_set <...>\n";
#   endif
#   if HAVE_GOOGLE_DENSE_HASH_SET
      cerr << "  google -- google::dense_hash_set <...>\n";
#   endif
      cerr << "  all -- all of the above in that order\n";

      cerr << "\nKnown benchmark names:\n";
      cerr << "  ie-int-(small|large|huge) -- insert/erase in a set of approx. 100, 10000\n"
           << "                               or 1000000 integers correspondingly\n";

      cerr << "\nLOOP-SIZE is optional and defaults to " << loop_size << ".\n";

      cerr << "\nNUM-TRIES is optional and defaults to " << num_tries << ". When greater than 1,\n"
           << "that many passes will be made and the best time reported.\n";

      return 2;
    }

  try
    {
      if (argc >= 4)
        do_parse (argv[3], loop_size);

      if (argc >= 5)
        do_parse (argv[4], num_tries);

      run_benchmark (argv[1], argv[2], loop_size, num_tries);
    }
  catch (error_exit&)
    {
      return 1;
    }

  return 0;
}


// Local variables:
// mode: c++
// c-basic-offset: 2
// indent-tabs-mode: nil
// fill-column: 90
// End:
